from typing import Any

__all__ = [
    "readBI",
    "writeBI",
    "overrideBI"
]

__BI = __builtins__


def readBI(name: str) -> Any:
    global __BI
    if isinstance(__BI, dict):
        return __BI[name]
    else:
        return getattr(__BI, name)


def writeBI(name: str, value: Any) -> None:
    global __BI
    if isinstance(__BI, dict):
        __BI[name] = value
    else:
        setattr(__BI, name, value)


def overrideBI(name: str, value: Any) -> Any:
    global __BI
    old = readBI(name)
    writeBI(name, value)
    return old



from typing import NoReturn

from ..error.BuiltinLockedError import BuiltinLockedError
from ...lang import lang


__all__ = [
    "LockedFunction",
    "LockedClass",
    "LockedSingleton"
]


def LockedFunction(*args, **kwargs) -> NoReturn:
    """
        This Function is LOCKED. No information available now
        这个函数被锁住了，没有可用的信息
        """
    raise BuiltinLockedError(lang["locked.function"])


class LockedClass:
    """
    This Class is LOCKED. No information available now
    这个类型被锁住了，没有可用的信息
    """

    def __init__(self, *args, **kwargs) -> NoReturn:
        raise BuiltinLockedError(lang["locked.class"])


class LockedSingletonType:
    """
    This Object is LOCKED. No information available now
    这个对象被锁住了，没有可用的信息
    """

    def __str__(self) -> NoReturn:
        raise BuiltinLockedError(lang["locked.class"])

    __repr__ = __str__


LockedSingleton = LockedSingletonType()



from typing import NoReturn, Any
from .Exceptions import YouShallNotUseIt
from .Lang import lang


def LockedFunction(*args, **kwargs) -> NoReturn:
    raise YouShallNotUseIt(lang.func.locked)


def BannedFunction(*args, **kwargs) -> NoReturn:
    raise YouShallNotUseIt(lang.func.banned)



from typing import Any

__all__ = [
    "Overrider"
]


class Overrider:
    BI: Any = __builtins__

    @staticmethod
    def read(name: str) -> Any:
        """
        platform implementation defined
        """
        pass

    @staticmethod
    def write(name: str, value: Any) -> None:
        """
        platform implementation defined
        """
        pass

    @staticmethod
    def override(name: str, value: Any) -> Any:
        origin = Overrider.read(name)
        Overrider.write(name, value)
        return origin


if isinstance(Overrider.BI, dict):
    def __read(name):
        return Overrider.BI[name]

    def __write(name, value):
        Overrider.BI[name] = value

else:
    def __read(name):
        return Overrider.BI.__getattribute__(name)


    def __write(name, value):
        Overrider.BI.__setattr__(name, value)


Overrider.read = __read
Overrider.write = __write



from ...lang import lang
from ..overrider import writeBI


class AnswerChecker:
    __printer = print
    __expected: str

    def __init__(self, answer):
        self.__expected = answer

    @property
    def expected(self):
        return self.__expected

    def print(self, *args, sep=' ', end='\n', file=None, flush=False):
        ans = sep.join(map(str, args)) + end
        if ans.strip() == self.__expected.strip():
            self.__printer(lang["answer.fit"])
            self.__printer(ans, file=file, flush=flush)
        else:
            self.__printer(lang["answer.unfit"])
            self.__printer(lang["answer.expected"], self.expected)
            self.__printer(lang["answer.actual"], ans)

    @staticmethod
    def override(answer: str):
        ac = AnswerChecker(answer)
        writeBI('print', ac)



from .Exceptions import YouShallNotUseIt
from .Lang import lang


class LockedType:
    def __init__(self, *args, **kwargs):
        raise YouShallNotUseIt(lang.type.locked)


class BannedType:
    def __init__(self, *args, **kwargs):
        raise YouShallNotUseIt(lang.type.banned)



from typing import Any

from .Locked import *
from ..overrider import overrideBI, writeBI


class __Cover:
    placeholder: Any
    __locked = {}

    @classmethod
    def lock(cls, name: str) -> None:
        cls.__locked[name] = overrideBI(name, cls.placeholder)

    @classmethod
    def unlock(cls, name: str) -> None:
        writeBI(name, cls.__locked[name])
        del cls.__locked[name]


class FunctionCover(__Cover):
    placeholder = LockedFunction


class ClassCover(__Cover):
    placeholder = LockedClass



TEXT = {
    "locked.function": "似乎这个函数已经被锁定了，通过关卡解锁吧。",
    "locked.class": "似乎这个类型已经被锁定了，通过关卡解锁吧。",

    "answer.fit": "恭喜，输出符合预期 ",
    "answer.unfit": "啊哦，似乎哪里出了问题。。。",
    "answer.expected": "预期结果：",
    "answer.actual": "你的结果："
}



from typing import Any
from .Lang import lang


class YouShallNotUseIt(Exception):
    
    def __init__(self, msg=None, *args):
        if msg is None:
            msg = lang.deny.default
        super().__init__(msg, *args)



from typing import Union
from .Overrider import Overrider
from ..covers import *

__all__ = [
    "ban"
]

BIF = type(input)


def ban(obj: Union["BIF", type]):
    if isinstance(obj, BIF):
        Overrider.write(obj.__name__, BannedFunction)

    elif isinstance(obj, type):
        if obj.__name__ not in __builtins__:
            raise ValueError("given obj is not in module builtins")
        Overrider.write(obj.__name__, BannedType)

    else:
        raise ValueError("Only support ban BIF or type")



from typing import Any, Union
from ..covers import *
from .Overrider import *


__all__ = [
    "FunctionLocker",
    "TypeLocker",
    "lock",
    "unlock"
]

BIF = type(input)
BIT = type


class __Locker:
    placeholder: Any
    __real = {}

    def __init__(self, placeholder: Any):
        self.placeholder = placeholder

    def lock(self, name: str):
        self.__real[name] = Overrider.override(name, self.placeholder)

    def unlock(self, name: str):
        if name not in self.__real:
            return
        Overrider.write(name, self.__real[name])
        del self.__real[name]


FunctionLocker = __Locker(LockedFunction)
TypeLocker = __Locker(LockedType)


def lock(obj: Union[BIF, type]):
    if isinstance(obj, BIF):
        FunctionLocker.lock(obj.__name__)

    elif isinstance(obj, type):
        if obj.__name__ not in __builtins__:
            raise ValueError("given obj is not in module builtins")
        TypeLocker.lock(obj.__name__)

    elif isinstance(obj, str):
        raise ValueError("a type or a function? param should be BIF or BIT, not types")

    elif obj == LockedFunction or obj == LockedType:
        return

    else:
        raise ValueError("Only support lock BIF or type")


def unlock(name: str):
    if not isinstance(name, str):
        raise ValueError("param shall be a string, try again.")
    FunctionLocker.unlock(name)
    TypeLocker.unlock(name)



from dataclasses import dataclass


@dataclass
class Lang:
    @dataclass
    class LangPack:
        locked: str
        lockDoc: str
        banned: str
        banDoc: str

    deny: str
    func: LangPack
    type: LangPack

    def applyDoc(self) -> None:
        """
        see __init__.py
        """
        pass


lang = Lang(
    "You Shall Not Use This",
    Lang.LangPack(
        "This Function is Locked",
        "This Function is Locked",
        "This Function is Banned",
        "This Function is Banned"
    ),
    Lang.LangPack(
        "This Type is Locked",
        "This Type is Locked",
        "This Type is Banned",
        "This Type is Banned"
    )
)



BIF = [
    'abs', 'all', 'any', 'bin',
    'callable', 'chr', 'compile',
    'dir', 'divmod',
    'eval', 'exec', 'exit', 'format',
    'frozenset', 'hash', 'hex', 'input',
    'isinstance', 'issubclass', 'iter',
    'len',  'max', 'min', 'next',
    'oct', 'open', 'ord', 'pow', 'print',
    'quit', 'repr', 'globals', 'locals',
    'round', 'sorted', 'sum',
    'vars']

BIT = [
    'bool', 'bytearray', 'bytes',
    'complex', 'dict', 'enumerate',
    'filter', 'float', 'int', 'list',
    'map', 'object', 'range',
    'staticmethod', 'classmethod', 'property',
    'reversed', 'set', 'slice', 'str', 'super',
    'tuple', 'type', 'zip'
]


__all__ = [
    "YouShallNotUseIt",
    "LockedType",
    "BannedType",
    "LockedFunction",
    "BannedFunction",
    "Lang", "lang"
]

from .Exceptions import YouShallNotUseIt
from .Types import LockedType, BannedType
from .Functions import LockedFunction, BannedFunction
from .Lang import Lang, lang


def applyDoc(self) -> None:
    pass
    LockedFunction.__doc__ = self.func.lockDoc
    BannedFunction.__doc__ = self.func.banDoc

    LockedType.__doc__ = self.type.lockDoc
    BannedType.__doc__ = self.type.banDoc


Lang.applyDoc = applyDoc

lang.applyDoc()
